use anyhow::{bail, Result};
use request::{Request, RequestKind};
use response::Response;
use std::{
    sync::{Arc, Mutex},
    thread,
};
use tokio::{
    io::{AsyncWriteExt as _, BufReader, BufWriter},
    net::{
        tcp::{OwnedReadHalf, OwnedWriteHalf},
        TcpListener, TcpStream,
    },
    select,
    sync::oneshot,
};

use crate::emulator::{EmulatorClient, EmulatorCommand, SimId};

mod request;
mod response;

pub struct GdbServer {
    sim_id: SimId,
    client: EmulatorClient,
    status: Arc<Mutex<GdbServerStatus>>,
    killer: Option<oneshot::Sender<()>>,
}

impl GdbServer {
    pub fn new(sim_id: SimId, client: EmulatorClient) -> Self {
        Self {
            sim_id,
            client,
            status: Arc::new(Mutex::new(GdbServerStatus::Stopped)),
            killer: None,
        }
    }

    pub fn status(&self) -> GdbServerStatus {
        self.status.lock().unwrap().clone()
    }

    pub fn start(&mut self, port: u16) {
        let sim_id = self.sim_id;
        let client = self.client.clone();
        let status = self.status.clone();
        let (tx, rx) = oneshot::channel();
        self.killer = Some(tx);
        thread::spawn(move || {
            tokio::runtime::Builder::new_current_thread()
                .enable_all()
                .build()
                .unwrap()
                .block_on(async move {
                    select! {
                        _ = run_server(sim_id, client, port, &status) => {}
                        _ = rx => {
                            *status.lock().unwrap() = GdbServerStatus::Stopped;
                        }
                    }
                })
        });
    }

    pub fn stop(&mut self) {
        if let Some(killer) = self.killer.take() {
            let _ = killer.send(());
        }
    }
}

async fn run_server(
    sim_id: SimId,
    client: EmulatorClient,
    port: u16,
    status: &Mutex<GdbServerStatus>,
) {
    let Some(stream) = try_connect(port, status).await else {
        return;
    };
    let connection = GdbConnection::new(sim_id, client, stream);
    match connection.run().await {
        Ok(()) => {
            *status.lock().unwrap() = GdbServerStatus::Stopped;
        }
        Err(error) => {
            *status.lock().unwrap() = GdbServerStatus::Error(error.to_string());
        }
    }
}

async fn try_connect(port: u16, status: &Mutex<GdbServerStatus>) -> Option<TcpStream> {
    *status.lock().unwrap() = GdbServerStatus::Connecting;
    let listener = match TcpListener::bind(("127.0.0.1", port)).await {
        Ok(l) => l,
        Err(err) => {
            *status.lock().unwrap() = GdbServerStatus::Error(err.to_string());
            return None;
        }
    };
    match listener.accept().await {
        Ok((stream, _)) => {
            *status.lock().unwrap() = GdbServerStatus::Running;
            Some(stream)
        }
        Err(err) => {
            *status.lock().unwrap() = GdbServerStatus::Error(err.to_string());
            None
        }
    }
}

#[derive(Clone)]
pub enum GdbServerStatus {
    Stopped,
    Connecting,
    Running,
    Error(String),
}

impl GdbServerStatus {
    pub fn running(&self) -> bool {
        matches!(self, Self::Connecting | Self::Running)
    }
}

struct GdbConnection {
    sim_id: SimId,
    client: EmulatorClient,
    stream_in: BufReader<OwnedReadHalf>,
    stream_out: BufWriter<OwnedWriteHalf>,
    ack_messages: bool,
    request_buf: Vec<u8>,
    response_buf: Option<Vec<u8>>,
}

impl GdbConnection {
    fn new(sim_id: SimId, client: EmulatorClient, stream: TcpStream) -> Self {
        let (rx, tx) = stream.into_split();
        Self {
            sim_id,
            client,
            stream_in: BufReader::new(rx),
            stream_out: BufWriter::new(tx),
            ack_messages: true,
            request_buf: vec![],
            response_buf: None,
        }
    }
    async fn run(mut self) -> Result<()> {
        println!("Connected for {}", self.sim_id);
        self.client.send_command(EmulatorCommand::Pause);
        loop {
            let mut req = Request::read(&mut self.stream_in, &mut self.request_buf).await?;
            println!("received {:?}", req);

            if req.kind == RequestKind::Signal {
                self.client.send_command(EmulatorCommand::Pause);
                let res = self
                    .response()
                    .write_str("T05;thread:p1.t1;threads:p1.t1;reason:trap;");
                self.send(res).await?;
                continue;
            }

            if req.match_str("QStartNoAckMode") {
                let res = self.response().write_str("OK");
                self.send(res).await?;
                self.ack_messages = false;
            } else if req.match_str("qSupported:") {
                let res = self
                    .response()
                    .write_str("multiprocess+;swbreak+;vContSupported+");
                self.send(res).await?;
            } else if req.match_str("QThreadSuffixSupported")
                || req.match_str("QListThreadsInStopReply")
                || req.match_str("QEnableErrorStrings")
            {
                let res = self.response().write_str("OK");
                self.send(res).await?;
            } else if req.match_str("qHostInfo") {
                let res = self.response().write_str(&format!(
                    "triple:{};endian:little;ptrsize:4;",
                    hex::encode("v810-unknown-vb")
                ));
                self.send(res).await?;
            } else if req.match_str("qProcessInfo") {
                let res = self.response().write_str(&format!(
                    "pid:1;triple:{};endian:little;ptrsize:4;",
                    hex::encode("v810-unknown-vb")
                ));
                self.send(res).await?;
            } else if req.match_str("vCont?") {
                let res = self.response().write_str("vCont;c;");
                self.send(res).await?;
            } else if req.match_str("qC") {
                // The v810 has no threads, so report that the "current thread" is 1.
                let res = self.response().write_str("QCp1.t1");
                self.send(res).await?;
            } else if req.match_str("qfThreadInfo") {
                let res = self.response().write_str("mp1.t1");
                self.send(res).await?;
            } else if req.match_str("qsThreadInfo") {
                let res = self.response().write_str("l");
                self.send(res).await?;
            } else if req.match_str("k") {
                bail!("debug process was killed");
            } else if req.match_str("?") {
                let res = self.response().write_str("T00;thread:p1.t1;threads:p1.t1;");
                self.send(res).await?;
            } else if req.match_str("c") || req.match_str("vCont;c:") {
                self.client.send_command(EmulatorCommand::Resume);
                // Don't send a response until we hit a breakpoint or get interrupted
            } else if req.match_str("p") {
                let Some(register) = req.match_hex::<u32>() else {
                    let res = self.response();
                    self.send(res).await?;
                    continue;
                };
                let res = self.response().write_hex(register);
                self.send(res).await?;
            } else if req.match_str("m") {
                let mut read_params = || {
                    let start = req.match_hex::<u32>()?;
                    if !req.match_str(",") {
                        return None;
                    };
                    let size = req.match_hex::<u32>()?;
                    Some((start, size))
                };
                let Some((start, size)) = read_params() else {
                    let res = self.response();
                    self.send(res).await?;
                    continue;
                };
                let mut res = self.response();
                for i in 0..size {
                    res = res.write_hex((start + i) as u8);
                }
                self.send(res).await?;
            } else {
                // unrecognized command
                let res = self.response();
                self.send(res).await?;
            }
        }
    }

    fn response(&mut self) -> Response {
        Response::new(
            self.response_buf.take().unwrap_or_default(),
            self.ack_messages,
        )
    }

    async fn send(&mut self, res: Response) -> std::io::Result<()> {
        let buffer = res.finish();
        println!("{:?}", std::str::from_utf8(&buffer));
        self.stream_out.write_all(&buffer).await?;
        self.response_buf = Some(buffer);
        self.stream_out.flush().await
    }
}
